# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from concurrent.futures import ThreadPoolExecutor
import json
import queue
import sys

import pyarrow
import pyarrow.flight
from azureml.featurestore._utils._constants import ON_THE_FLY_ENTITY_KEY, PACKAGE_NAME
from azureml.featurestore._utils.utils import _build_logger
from azureml.featurestore.online._online_feature_getter_v2 import OnlineFeatureGetterV2

from azure.ai.ml._telemetry.activity import ActivityType, log_activity
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, MlException, ValidationException
from azure.core.credentials import AccessToken

package_logger = None


def _get_logger():
    '''Get the logger
    Calling this function anywhere in the file will get the same logger 
    '''
    global package_logger
    if package_logger is None:
        package_logger = _build_logger(__name__)
    return package_logger


class AuthTokenCredential(object):
    def __init__(self, token_dict):
        self.tokens = token_dict

    def get_token(self, *scopes, **kwargs):
        if len(scopes) != 1:
            msg = "This credential requires exactly one scope per token request."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.IDENTITY,
                error_category=ErrorCategory.USER_ERROR,
            )

        token = self.tokens[scopes[0]]
        return AccessToken(token["token"], token["expires_on"])


class FlightFeatureRetrievalServer(pyarrow.flight.FlightServerBase):
    def __init__(self, location, credential, feature_uris, on_the_fly_feature_set_uris):
        self.log_queue = queue.Queue()
        self.logger = _get_logger()

        with log_activity(
            self.logger,
            f"{PACKAGE_NAME}->FlightFeatureRetrievalServer->Init",
            ActivityType.INTERNALCALL,
            { "feature_count": len(feature_uris) }
        ) as activity:
            print("Initializing feature getter...", flush=True)
            self.online_feature_getter = OnlineFeatureGetterV2(credential, feature_uris, on_the_fly_feature_set_uris)
            print("Finished feature getter initiailization! Starting server...", flush=True)
            super(FlightFeatureRetrievalServer, self).__init__(location)
            print("Started gRPC server!\n----\n", flush=True)
            activity.activity_info["redis_database_count"] = len(self.online_feature_getter.redis_clients)
        

    def do_exchange(self, context, descriptor, reader, writer):
        """Write data to a flight.
        Applications should override this method to implement their
        own behavior. The default method raises a NotImplementedError.
        Parameters
        ----------
        context : ServerCallContext
            Common contextual information.
        descriptor : FlightDescriptor
            The descriptor for the flight provided by the client.
        reader : MetadataRecordBatchReader
            A reader for data uploaded by the client.
        writer : MetadataRecordBatchWriter
            A writer to send responses to the client.
        """
        try: 
            # Get feature list from descriptor
            scenario = descriptor.path[0].decode("utf-8")

            if scenario == "online":
                feature_getter = self.online_feature_getter.get_online_features
            elif scenario.startswith("offline:"):
                raise NotImplementedError("Offline feature data retrieval over grpc is not yet supported.")
            else:
                raise NotImplementedError(f"Unsupported scenario: {scenario}")

            if descriptor.path[1].decode("utf-8") == ON_THE_FLY_ENTITY_KEY:
                on_the_fly_entities = descriptor.path[2].decode("utf-8")
                feature_uris = [path.decode("utf-8") for path in descriptor.path[3:]]
            else:
                on_the_fly_entities = None
                feature_uris = [path.decode("utf-8") for path in descriptor.path[1:]]

            # Get observations dataframe from request
            observation_df = reader.read_all()
            features_df = feature_getter(feature_uris, observation_df, on_the_fly_entities=on_the_fly_entities)

            writer.begin(features_df.schema)
            writer.write_table(features_df)
            writer.close()
        except Exception as ex:
            _log_failure(self.logger, ex)
            raise
        else:
            self.log_queue.put(
                (features_df.num_rows, features_df.num_columns, features_df.nbytes)
            )
            

def _log_success(logger, num_rows, num_cols, num_bytes):
    """Log a successful feature data retrieval event.
    This method should *never* throw."""
    try:
        with log_activity(
            logger,
            f"{PACKAGE_NAME}->FlightFeatureRetrievalServer->DoExchange->Success",
            ActivityType.INTERNALCALL,
            {
                "feature_data_rows": num_rows,
                "feature_data_columns": num_cols,
                "feature_data_bytes": num_bytes,
            }
        ):
            pass
    except:
        pass


def _log_failure(logger, exception):
    """Log an unsuccessful feature data retrieval event.
    This method should *never* throw."""
    try:
        if isinstance(exception, MlException):
            log_activity(
                logger,
                f"{PACKAGE_NAME}->FlightFeatureRetrievalServer->DoExchange->Error",
                ActivityType.INTERNALCALL,
                {
                    "exception": f"{type(exception).__name__}: {exception.no_personal_data_message}",
                },
            )
        else:
            log_activity(
                logger,
                f"{PACKAGE_NAME}->FlightFeatureRetrievalServer->DoExchange->Error",
                ActivityType.INTERNALCALL,
                {
                    "exception": f"{type(exception).__name__}: {exception}",
                },
            )
    except:
        pass


def sentinel(server):
    # Wait as long as stdin is open.
    for line in sys.stdin:
        pass
    
    # stdin was closed - the parent process is likely dead.
    # Emit telemetry to appinsights
    with log_activity(
            _get_logger(),
            f"{PACKAGE_NAME}->FlightFeatureRetrievalServer->Shutdown",
            ActivityType.INTERNALCALL
        ):
        pass

    # Flush logs
    sys.stdout.flush()
    sys.stderr.flush()
    
    # Shut the server down
    server.shutdown()
    print(f"\n----\nServer shuts down.", flush=True)

     # Signal the log worker to exit
    server.log_queue.put(None)


def log_worker(server):
    while True:
        log = server.log_queue.get() # Wait here until an item is available; Since it is a blocking operation, no try-catch block is needed.
        if log is None:
            break
        num_rows, num_cols, num_bytes = log
        _log_success(server.logger, num_rows, num_cols, num_bytes)

    print("Log queue has been exhausted.", flush=True)


def main(location, credential, feature_uris, on_the_fly_feature_sets):
    server = FlightFeatureRetrievalServer(location, credential, feature_uris, on_the_fly_feature_sets)
    with ThreadPoolExecutor(max_workers=3) as executor:
        executor.submit(sentinel, server)
        executor.submit(log_worker, server)

    server.serve()


if __name__ == "__main__":
    # Read initialization params from stdin
    initialization_params = json.loads(sys.stdin.readline())
    location = initialization_params["location"]
    feature_uris = initialization_params["features"]
    credential = AuthTokenCredential(initialization_params["tokens"])
    on_the_fly_feature_sets = initialization_params["on_the_fly_feature_sets"]

    main(location, credential, feature_uris, on_the_fly_feature_sets)
